/* BVI  -  Binary Visual Editor
 *
 * 1996-02-28  V 1.00  created
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * Copyright 1996 by Gerhard Buergmann (G.Buergmann@nl.cis.philips.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifdef __MSDOS__
#	include "doscur.h"
#   include <alloc.h>
#else
#	include <curses.h>
#endif

#ifdef MINICURSES
#	define NOGETMAXYX
#endif

#include "set.h"
#include "bvi.h"


char	*version  = "Version 1.00 by Gerhard Buergmann";

char contrd[][4] = {"NUL", " ^A", " ^B", " ^C", " ^D", " ^E", " ^F", "BEL",
					" BS", "TAB", " NL", "HOM", "CLR", " CR", " ^N", " ^O",
					" ^P", " ^Q", " ^R", " ^S", " ^T", " ^U", " ^V", " ^W",
					" ^X", " ^Y", " ^Z", "ESC", " FS", " GS", " RS", " US" };
char contru[][4] = {"NUL", "SOH", "STX", "ETX", "EOT", "ENQ", "ACK", "BEL",
					" BS", " HT", " NL", " VT", " NP", " CR", " SO", " SI",
					"DLE", "DC1", "DC2", "DC3", "DC4", "NAK", "SYN", "ETB",
					"CAN", " EM", "SUB", "ESC", " FS", " GS", " RS", " US" };

int		new = 0;
int		loc;
int		maxx, maxy, x, xx, y;
int		screen, status;
unsigned int nr;
long	size;
PTR		mem = NULL;
PTR		curpos;
PTR		maxpos;
PTR		pagepos;
PTR		spos;
char	*name = NULL;
char	linbuf[256];
char	tmpbuf[10];
char	string[MAXCMD];
char	cmdstr[MAXCMD + 1];
FILE	*Ausgabe_Datei;
int		edits = 0;
int		count;
int		Anzahl, Anzahl3;
long	filesize, memsize, undosize;
char	*shell;

char	**files;		/* list of input files */
int		numfiles;		/* number of input files */
int		curfile;		/* number of the current file */

int		precount = 0;	/* number preceding command */
int		arrnum = 0;
char	numarr[80];		/* string for collecting number */
char	progname[15];

PTR		current;
PTR		last_motion;
PTR		undo_start;
PTR		current_start;
int		undo_count;
char	undo_buf[BUFFER];
int		yanked = 0;
char	yank_buf[BUFFER];
PTR		markbuf[26];

char	*strdup();

void
usage()
{
	fprintf(stderr, "Usage: %s [-R] [-c command] file ...\n", progname);
	exit(1);
}


int
main(argc, argv)
	int argc;
	char *argv[];
{
	char	temp;
	int		ch;
	int		lflag;
	int		mark, scrolly;
	int		n = 1, wrstat = 1;
	long	inaddr, tempsize;
	char	line[MAXCMD];
	char	*poi;

	poi = strrchr(argv[0], DELIM);

	if (poi) strcpy(progname, ++poi);
		else strcpy(progname, argv[0]);
	strtok(progname, ".");

	if (!strcasecmp(progname, "bview")) {
		params[P_RO].flags |= P_CHANGED;
		P(P_RO) = TRUE;
	} else if (!strcasecmp(progname, "bvedit")) {

	}

	while (n < argc) {
		switch (argv[n][0]) {
		case '-':
			if (argv[n][1] == 'R') {
				params[P_RO].flags |= P_CHANGED;
				P(P_RO) = TRUE;
			} else if (argv[n][1] == 'c') {
				if (argv[n + 1] == NULL) {
					usage();
				} else {
					sprintf(string, "%s\n", argv[++n]);
					stuffin(string);
				}
			} else if (argv[n][1] == 'w') {
				if (argv[n][2] == '\0') {
					usage();
				} else {
					params[P_LI].flags |= P_CHANGED;
					P(P_LI) = atoi(argv[n] + 2);
				}
			} else usage();
			n++;
			break;
		case '+':			/* +/pat */
			if (argv[n][1] == '/') {
					sprintf(string, "%s\n", &argv[n][1]);
					stuffin(string);
			} else
				usage();
			n++;
			break;
		default:			/* must be a file name */
			name = strdup(argv[n]);
			files = &(argv[n]);
			numfiles = argc - n;
			n = argc;
			break;
		}
	}
	if (numfiles > 1) printf("%d files to edit\n", numfiles);
	curfile = 0;

	/****** Initialisation of curses ******/
	initscr();
	attrset(A_NORMAL);

#ifdef NOGETMAXYX
	maxx = ((stdscr)->_maxx);
	maxy = ((stdscr)->_maxy);
#else
	getmaxyx(stdscr, maxy, maxx);
#endif

	if (params[P_LI].flags & P_CHANGED) maxy = P(P_LI);
	scrolly = maxy / 2;
	P(P_SS) = scrolly;
	P(P_LI) = maxy;
	maxy--;
	keypad(stdscr, TRUE);

	scrollok(stdscr, TRUE);

	nonl();
	cbreak();
	noecho();

	Anzahl = ((maxx - 9) / 16) * 4;
	maxx = Anzahl * 4 + 9;
	Anzahl3 = Anzahl * 3;
	status = Anzahl3 + Anzahl - 17;
	screen = Anzahl * (maxy - 1);

	signal(SIGINT, SIG_IGN);

	bvi_init(argv[0]);

	filesize = load(name);

	do {
		current = (PTR)(pagepos + y * Anzahl + xpos());
		if (wrstat) statpos();
		wrstat = 1;
		setcur();
		ch = vgetc();
		if (ch != '0') {
			while (ch >= '0' && ch <= '9') {
				numarr[arrnum++] = ch;
				ch = vgetc();
				}
			}
		numarr[arrnum] = '\0';
		precount = atoi(numarr);
		lflag = arrnum = 0;

		switch (ch) {
		case '0':	x = 8; loc = HEX; break;
		case '^':	x = 8 + Anzahl3; loc = ASCII; break;
		case '$':	x = 7 + Anzahl3 + Anzahl; loc = ASCII; break;
		case '\t':  toggle(); break;

		case KEY_HOME:
		case 'H':	if (precount) {
						y = --precount;
						if (y > maxy - 1) {
							scrolldown(y - maxy + 1);
							y = maxy - 1; }
					} else y = 0;
					if (loc == HEX) x = 8;
						else	x = 8 + Anzahl3;
					break;
		case 'M':	y = maxy / 2;
					if ((PTR)(pagepos + screen) > maxpos)
						y = (int)(maxpos - pagepos) / Anzahl / 2;
					if (loc == HEX) x = 8;
						else	x = 8 + Anzahl3;
					break;
		case KEY_LL:
		case 'L':   if (!precount) precount++;
					n = maxy - 1;
					if ((PTR)((pagepos + screen)) > maxpos)
						n = (int)(maxpos - pagepos) / Anzahl;
					if (precount < n) y = n + 1 - precount;
					if (loc == HEX) x = 8;
						else	x = 8 + Anzahl3;
					break;
		case CTRL('H'):
		case KEY_BACKSPACE:
		case KEY_LEFT:
		case 'h':	do {
						if (x > 10 && x < (Anzahl3 + 9))  x -= 3;
							else   if (x > (Anzahl3 + 6))  x--;
					} while (--precount > 0);
					if (x < 8 + Anzahl3) loc = HEX;
						else loc = ASCII;
					break;
		case ' ':
		case KEY_RIGHT:
		case 'l':	do {
						if (x < (Anzahl3 + 6))  x += 3;
							else if (x > (Anzahl3 + 3)
								&& x < (Anzahl3 + 7 + Anzahl))
									x++;
					} while (--precount > 0);
					if (x < 8 + Anzahl3) loc = HEX;
						else loc = ASCII;
					break;
		case '-':
		case KEY_UP :
		case 'k':	do {
						if (y > 0) y--;
							else scrollup(1);
					} while(--precount > 0);
					break;
		case 'j':
		case '+':
		case KEY_DOWN:
		case CR:	do {
						if ((PTR)((pagepos + (y + 1) * Anzahl)) > maxpos) break;
						if (y < (maxy - 1))	y++;
						else scrolldown(1);
					} while(--precount > 0);
					break;
		case '|':   if (!precount) break;
					if (loc == ASCII) x = 7 + Anzahl3 + precount;
						else x = 5 + 3 * precount;
					if (x > 7 + Anzahl3 + Anzahl) {
						x = 7 + Anzahl3 + Anzahl;
						loc = ASCII; }
					break;
		case ':' :	clearstr();
					addch(ch);
					refresh();
					getcmdstr(cmdstr, 1);
					if (strlen(cmdstr))
						docmdline(cmdstr);
					break;
		case CTRL('B'):
		case KEY_PPAGE: 	/**** Previous Page ****/
					if (mem <= (PTR)(pagepos - screen))	pagepos -= screen;
						else	pagepos = mem;
					repaint();
					break;
		case CTRL('D'):
					if (precount) scrolly = precount;
					scrolldown(scrolly);
					break;
		case CTRL('U'):
					if (precount) scrolly = precount;
					scrollup(scrolly);
					break;
		case CTRL('E'):
					if (y > 0) y--;
					scrolldown(1);
					break;
		case CTRL('F'):
		case KEY_NPAGE: 	/**** Next Page *****/
					if (maxpos >= (PTR)(pagepos + screen))  {
						pagepos += screen;
						repaint(); }
					break;
		case CTRL('G'):
					fileinfo(name);
					wrstat = 0;
					break;
		case CTRL('L'):   	/*** REDRAW SCREEN ***/
					refresh();
					break;
		case CTRL('Y'):
					if (y < maxy) y++;
					scrollup(1);
					break;
		case 'A':   smsg("APPEND MODE");
					setpage((PTR)(mem + filesize - 1L));
					cur_forw();
					setcur();
					undosize = filesize;
					edit(-1);
					break;
		case 'B':
		case 'b':	setpage(backsearch(current, ch));
					break;
		case 'D':	edits = TRUNC;
					undosize = filesize;
					filesize = pagepos - mem + y * Anzahl + xpos();
					maxpos = (PTR)(mem + filesize);
					if (filesize == 0L) {
						emsg("No bytes@in the buffer");
					} else	cur_back();
					repaint();
					break;
		case 'e':	setpage(end_word(current));
					break;
		case 'G':   last_motion = current;
					if (precount)	setpage((PTR)(mem + (long)precount - 1L));
						else	setpage((PTR)(mem + filesize - 1L));
					break;
		case 'g':	last_motion = current;
					msg("Goto Hex Address: ");
					refresh();
					getcmdstr(cmdstr, 19);
					sscanf(cmdstr, "%lx", &inaddr);
					if (inaddr < P(P_OF)) break;
					inaddr -= P(P_OF);
					if (inaddr < filesize)  {
						setpage(mem + inaddr);
					} else {
						if (filesize == 0L) break;
						sprintf(string, "Max. address of current file : %06lX",
							filesize - 1L + P(P_OF));
						emsg(string);
					}
					break;
		case 'i':
		case 'I':	emsg("No INSERT MODE, use 'r', 'R' or 'A' !");
					break;
		case 'f':
		case 'F':	if (loc == HEX) toggle();
					searching(ch, line, SEARCH);
					break;
		case '?':
		case '/':	/**** Search String ****/
		case '%':
		case '\\':  clearstr();
					addch(ch);
					refresh();
					if (getcmdstr(line, 1)) break;
					if (line[0] == '\0') break;
		case 'n': 		/**** Search Next ****/
		case 'N':   last_motion = current;
					searching(ch, line, SEARCH);
					break;
		case 'm':	mark = vgetc();
					if (mark < 'a' || mark > 'z') beep();
					else  markbuf[mark - 'a'] = current;
					break;
		case '\'':
		case '`':   if ((ch == '`' && loc == ASCII) || (ch == '\'' && loc == HEX))
						toggle();
					mark = vgetc();
					if (mark == '`' || mark == '\'') {
						setpage(last_motion);
						last_motion = current;
					} else {
						if (mark < 'a' || mark > 'z') {
							beep(); break;
						} else if (markbuf[mark - 'a'] == NULL) {
								beep(); break;
								}
						setpage(markbuf[mark - 'a']);
					}
					break;
		case 'p':	undo_start = current;
					undo_count = yanked;
					for (n = 0; n < yanked; n++) {
						temp = undo_start[n];
						undo_start[n] = yank_buf[n];
						undo_buf[n] = temp;
						}
					edits = EDIT;
					repaint();
					break;
		case 'P':   if (filesize + yanked > memsize) {
						if (enlarge(yanked + 100L))	break;
						}
					setpage(mem + filesize - 1L);
					cur_forw();
					setcur();
					undosize = filesize;
					memcpy(mem + filesize, yank_buf, yanked);
					edits = TRUNC;
					filesize += yanked;
					repaint();
					break;
		case 'r':   if (filesize == 0L) break;
					if (current >= maxpos) break;
					smsg("REPLACE 1 CHAR");
					undo_start = current;
					if (!precount) precount++;
					undo_count = edit(precount);
					lflag++;
					break;
		case 'R':   if (filesize == 0L) break;
					if (current >= maxpos) break;
					smsg("REPLACE MODE");
					undo_start = current;
					undo_count = edit(0);
					lflag++;
					break;
		case 'u':   switch (edits) {
					case EDIT:
						for (n = 0; n < undo_count; n++) {
							temp = undo_start[n];
							undo_start[n] = undo_buf[n];
							undo_buf[n] = temp;
							}
						setpage(undo_start);
						break;
					case TRUNC:
						tempsize = filesize;
						filesize = undosize;
						undosize = tempsize;
						setpage(mem + filesize - 1L);
						break;
					}
					repaint();
					break;
		case 'W':
		case 'w':	loc = ASCII;
					setpage(wordsearch(current, ch));
					break;
		case 'y':   if (!precount) precount++;
					if (precount > BUFFER) {
						beep();
						break;
					}
					yanked = precount;
					memcpy(yank_buf, current, precount);
					break;
		case 'z':
					do_z(vgetc());
					break;
		case 'Z':   if (vgetc() == 'Z') do_exit();
						else beep();
					break;
		case '.':   if (!undo_count) {
						beep();
						break; }
					current_start = current;
					for (n = 0; n < undo_count; n++) {
						undo_buf[n] = current_start[n];
						current_start[n] = undo_start[n];
						}
					undo_start = current_start;
					repaint();
					break;
		default :   flushinp();
					beep();
		}
		if (lflag) lineout();
	} while (1);
}


void
do_z(mode)
	int	mode;
{
	switch (mode) {
	case '.':	while (y != maxy / 2) {
					if (y > maxy / 2) {
						pagepos += Anzahl;
						y--;
					} else {
						if (pagepos == mem) break;
						pagepos -= Anzahl;
						y++;
					}
				}
				break;
	case '-':	while (y < maxy - 1) {
					if (pagepos == mem) break;
					pagepos -= Anzahl;
					y++;
					}
				break;
	case '\0':
	case '\n':
	case '\r':	while (y > 0) {
					y--;
					pagepos += Anzahl;
				}
				break;
	default :   beep();
				break;
	}
	repaint();
}


void
scrolldown(lines)
	int	lines;
{
	while (lines--) {
		if (maxpos >= (pagepos + Anzahl)) pagepos += Anzahl;
			else { beep(); lines = 0; }
		repaint();
		refresh();
	}
}


void
scrollup(lines)
	int	lines;
{
	while (lines--) {
		if (mem <= (PTR)(pagepos - Anzahl)) pagepos -= Anzahl;
			else { beep(); lines = 0; }
		repaint();
		refresh();
	}
}


int
xpos()
{
	if (loc == HEX) return((x - 8) / 3);
		else   return(x - 8 - Anzahl3);
}


void
toggle()
{
	if (loc == HEX) {
		x = xpos() + 8 + Anzahl3;
		loc = ASCII;
	}	else	{
		x = xpos() * 3 + 8;
		loc = HEX;
	}
}


/* mode: -1 append
 *        0 replace one or more different char
 *       >0 replace n char with one
 */
edit(mode)
	int	mode;
{
	int		ch, ch1;
	int		count = 0;
	long	origsize;

	origsize = filesize;
	while((ch = vgetc()) != 27) {
		ch &= 0xff;
		if (ch == '\t') {
			toggle();
			setcur();
			continue;
		}
		if (ch == KEY_BACKSPACE || ch == CTRL('H')) {
			if (count > 0) {
				count--;
				if (filesize > origsize) {
					filesize--;
					maxpos--; }
				cur_back();
				setcur();
			} else beep();
			continue;
			}
		if (loc == HEX) {
			if (isxdigit(ch)) {
				mvaddch(y, x + 1, ' ');
				mvaddch(y, x, ch);
				while (!isxdigit(ch1 = vgetc() & 0xff)) beep();
				mvaddch(y, x + 1, ch1);
				sprintf(tmpbuf, "%c%c", ch, ch1);
				sscanf(tmpbuf, "%2x", &ch);
			} else {
				beep();
				goto wrong;
			}
		} else {			/*** ASCII - Bereich ***/
			if (isprint(ch)) {
				mvaddch(y, x, ch);
			} else {
				beep();
				goto wrong;
			}
		}
		curpos = pagepos + y * Anzahl + xpos();
		current = curpos + 1L;
		undo_buf[count++] = *curpos;
		if (curpos == maxpos) {
			edits = TRUNC;
			maxpos++;
			filesize++;
		} else edits = EDIT;
		*curpos = (char)ch;
		cur_forw();
		statpos();
		lineout();
		setcur();

		if (filesize > memsize - 2) {
			if (enlarge(100L)) break;
		}

		if (mode > -1 && curpos == maxpos - 1) break;
		if (mode > 0) {
			while (--mode) {
				if (++curpos < maxpos) {
					undo_buf[count++] = *curpos;
					*curpos = (char)ch;
					cur_forw();
					statpos();
					lineout();
					}
				}
			break;
		}
wrong:
	continue;
	}
	cur_back();
	setcur();
	smsg("");
	return(count);
}


void
setcur()
{
	move(y, x);
	refresh();
}


void
statpos()	/************* display current position *************/
{
	unsigned char  Char1;
	long	curpos;
	char	string[30], str[4];

	if (!P(P_MO)) return;
	curpos = current - mem;
	if (curpos >= filesize) {
		mvaddstr(maxy, status, "                         ");
		return; }
	Char1 = *(mem + curpos);
	if (isprint(Char1)) sprintf(str, "'%c'", Char1);
	else if (Char1 < 32) {
			if (P(P_US))	strcpy(str, contru[Char1]);
				else        strcpy(str, contrd[Char1]);
		} else strcpy(str, "   ");

	sprintf(string, "%06lX  \\%03o 0x%02X %3d %3s",
				curpos + P(P_OF), Char1, Char1, Char1, str);
	attrset(A_BOLD);
	mvaddstr(maxy, status, string);
	attrset(A_NORMAL);
}


int
lineout()   /************* display a line ************************/
{
	unsigned char Zeichen;
	long	Adresse;
	int		Ausgabe_Pos;

	Adresse = pagepos - mem + y * Anzahl;
	if (Adresse > filesize) strcpy(linbuf, "~       ");
	else {
		sprintf(linbuf, "%06lX  ", Adresse + P(P_OF));
		*string = '\0'; }
	for (Ausgabe_Pos = 0; Ausgabe_Pos < Anzahl; Ausgabe_Pos++) {
		if (Adresse >= filesize) {
			sprintf(tmpbuf, "   ");
			Zeichen = ' ';
		} else {
			Zeichen = *(mem + Adresse);
			sprintf(tmpbuf, "%02X ", Zeichen);
			++Adresse;
		}
		strcat(linbuf, tmpbuf);
		if (isprint(Zeichen))
			*(string + Ausgabe_Pos) = Zeichen;
		else
			*(string + Ausgabe_Pos) = '.';
	}
	*(string + Anzahl) = '\0';
	strcat(linbuf, string);
	mvaddstr(y, 0, linbuf);
	move(y, x);
	return(0);
}


void
new_screen()
{
	screen = Anzahl * (maxy - 1);
	clear();
	repaint();
}


void
repaint()	/***** redraw screen *********************/
{
	int	save_y;

	save_y = y;
	for (y = 0; y < maxy; y++) lineout();
	y = save_y;
}



void
setpage(addr)	/******* display an arbitrary address on screen *******/
	PTR		addr;
{
	PTR		actpos;

	actpos = (PTR)(((addr - mem) / Anzahl) * Anzahl + mem);
	if ((addr >= pagepos) && ((addr - pagepos) < screen)) {
		y = (addr - pagepos) / Anzahl;
		if (loc == HEX)
			x = 8 + ((addr - pagepos) - y * Anzahl) * 3;
		else
			x = 8 + Anzahl3  + ((addr - pagepos) - y * Anzahl);
	} else {
		pagepos = (PTR)(actpos - (long)(Anzahl * (maxy / 2)));
		if (pagepos < mem) pagepos = mem;
		y = (addr - pagepos) / Anzahl;
		if (loc == HEX)
			x = 8 + ((addr - pagepos) - y * Anzahl) * 3;
		else
			x = 8 + Anzahl3 + ((addr - pagepos) - y * Anzahl);
		repaint();
	}
}


void
cur_forw()
{
	if (loc == ASCII) {
		if (x < 7 + Anzahl3 + Anzahl) { x++ ; return; }
			else x = 8 + Anzahl3;
	} else {
		if (x < 5 + Anzahl3) { x += 3; return; }
			else x = 8;
	}
	statpos();
	lineout();
	if (y < maxy - 1) { y++; return; }
		else {
			if (pagepos < (PTR)(mem + filesize)) {
					pagepos += Anzahl;
					repaint();
					return; }
				else beep();
		}
}


void
cur_back()
{
	if (loc == ASCII) {
		if (x > 8 + Anzahl3) { x-- ; return; }
			else x = 7 + Anzahl3 + Anzahl;
	} else {
		if (x > 10) { x -= 3; return; }
			else x = 8 + Anzahl3;
	}
	statpos();
	lineout();
	if (y > 0) { y--; return; }
		else {
			if (pagepos > mem) {
					pagepos -= Anzahl;
					repaint();
					return; }
				else beep();
		}
}


void
fileinfo(fname)
	char	*fname;
{
	long	actpos;

	if (!fname) sprintf(string, "No file  %ld bytes", filesize);
	else {
		actpos = (pagepos + y * Anzahl + xpos()) - mem + 1L;
		sprintf(string, "\"%s\" %s%sbyte %ld of %ld --%ld%%--", fname,
			P(P_RO) ? "[Read only] " : "",
			edits ? "[Modified] " : "",
			actpos, filesize, actpos * 100L / filesize);
	}
	msg(string);
}


void
quit()      /********************* End ******************************/
{
	move(maxy, 0);
	endwin();
	printf("\n%s\n", version);
	exit(0);
}


