/* COMM.C - Routines to parse and execute "command line" commands, such as
 * search or colon commands.
 *
 * 1996-01-16 created;
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * Copyright 1996 by Gerhard Buergmann (G.Buergmann@nl.cis.philips.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */


#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>

#ifdef __MSDOS__
#	include "doscur.h"
#	include <alloc.h>
#	include <io.h>
#else
#	include <curses.h>
#endif

#include "bvi.h"
#include "set.h"

#define cmdlng(a,b)		(len <= a && len >= b)
#define strsave strdup
#define	CMDSZ	100		/* size of the command buffer */
#define	RBSIZE	1024

static	char	*altfile = NULL;	/* alternate file */
static	struct	stat buf;

char	*nowrtmsg  = "No write@since last change (:%s! overrides)";
char	*morefiles = "more files@to edit";
char	*ambigous  = "Ambigous|Too many file names";
char	*extra     = "Extra chars|Extra characters at end of command";
int		repl_count;

extern	char	*name;			/* actual filename */
extern	char	**files;		/* used for "next" and "rewind" */
extern	int		numfiles, curfile;
extern	int		errno;

static char getcbuff[RBSIZE];
static char *getcnext = NULL;

static char oldbuf[CMDSZ];		/** for :!! command **/


void
stuffin(s)
char	*s;
{
	if (s == NULL) {		/* clear the stuff buffer */
		getcnext = NULL;
		return;
	}

	if (getcnext == NULL) {
		strcpy(getcbuff, s);
		getcnext = getcbuff;
	} else
		strcat(getcbuff, s);
}


int
vgetc()
{
	if ( getcnext != NULL ) {
		int nextc = *getcnext++;
		if ( getcnext == NULL ) {
			*getcbuff = '\0';
			getcnext = NULL;
		}
		return(nextc);
	}
	return getch();
}



void
clear_marks()
{
	int n;

	for (n = 0; n < 26; markbuf[n++] = NULL);
	undo_count = 0;
	last_motion = mem;
}


/*
 * docmdline() - handle a colon command
 *
 * Handles a colon command received interactively by getcmdln() or from
 * the environment variable "BVIINIT" (or eventually .bvirc).
 */
void
docmdline(cmdline)
	char	*cmdline;
{
	char	buff[CMDSZ];
	char	cmdbuf[CMDSZ];
	char	argbuf[CMDSZ];
	char	*cmd;
	register char	*p;
	int		len, n;
	int		ok;
	int		force = 0;
	int		c_argc = 0;
	char	*c_argv[9];

	if (cmdline == NULL) return;
	if (*cmdline == '"') return; 	/** comment **/
	if (strlen(cmdline) > CMDSZ - 2) {
		emsg("Command too long");
		return; }

	strcpy(buff, cmdline);

	/* skip any initial white space */
	for (cmd = buff; *cmd != '\0' && isspace(*cmd) ;cmd++);

	while ((p = strchr(cmd, '#')) != NULL && *(p - 1) != '\\') {
					/* change '#' to Altname */
		if (altfile == NULL) {
			emsg("No alternate filename@to substitute for #");
			return;
		}
		*p = '\0';
		strcpy (cmdbuf, cmd);
		strcat (cmdbuf, altfile);
		strcat (cmdbuf, p + 1);
		strcpy(cmd, cmdbuf);
		msg(cmd);			/*repeat */
	}

	strcpy(cmdbuf, cmd);	/* save the unmodified command */

	if (*cmd == '!') {
		if (*(cmdbuf + 1) == '\0') {
			emsg("Incomplete shell escape command@- use 'shell' to get a shell");
			return;
			}
		if (*(cmdbuf + 1) == '!') {
			if (oldbuf[0] != '\0') {
				strcpy(cmdbuf + 1, oldbuf);
				msg(oldbuf);
			} else {
				emsg("No previous command@to substitute for !");
				return;
			}
		} else	sprintf(oldbuf, "\n%s\n", cmdbuf + 1);

		if (P(P_AW)) {
			save(name);
			edits = 0;
		}
		if (edits) msg("[No write]|[No write since last change]");
		savetty();
		endwin();
		system(cmdbuf + 1);
		fprintf(stderr, "[Hit return to continue]");
		getchar();
		doupdate();
		repaint();
		clearstr();
		return;
	}

	if (*cmd == 's' && strchr("/\\?%", *(cmd + 1))) {
		repl_count = 0;
		searching(*(cmdbuf + 1), cmdbuf + 2, REPLACE);
		repaint();
		if (!repl_count) {
			emsg("Fail|Substitute pattern matching failed");
		} else if (repl_count > 1) {
			sprintf(string, "%d subs|%d substitutions", repl_count, repl_count);
			msg(string);
		}
		return;
	}

	c_argc = 0;
	cmd = strtok(cmdbuf, " \t");
	while ((c_argv[c_argc] = strtok(NULL, " \t")) != NULL) c_argc++;

	len = strlen(cmd);
	if (cmd[len - 1] == '!') {
		force = 1;
		len--; }

	if (!strncmp("set", cmd, len) && cmdlng(3, 2)) {
		if (c_argc == 0) doset(NULL);
		else
			for (n = 0; n < c_argc; n++) {
				if (doset(c_argv[n])) return;
				}
	} else if (!strncmp("args", cmd, len) && cmdlng(4, 2)) {
		if (c_argc == 0) {
			strcpy(string, "");
			for (n = 0; n < numfiles; n++) {
				if (n == curfile) strcat(string, "[");
				strcat(string, files[n]);
				if (n == curfile) strcat(string, "]");
				strcat(string, " ");
			}
			msg(string);
		} else {
			/* this is the place to set new arguments **********/
			/*
			name = strdup(c_argv[0]);
			files = &(c_argv[0]);
			numfiles = c_argc;
			*/
		}
	} else if (!strncmp("source", cmd, len) && cmdlng(6, 2)) {
    	if (c_argc > 1) { emsg(ambigous); return; }
		if (c_argc == 0) { emsg("Missing filename"); return; }
		if (read_rc(c_argv[0])) sysemsg(c_argv[0]);
		refresh();
	} else if (!strncmp("cd", cmd, len) && cmdlng(2, 2)) {
    	if (c_argc > 1) { emsg(ambigous); return; }
		if (c_argc == 0) {
			sprintf(string, "%c", DELIM);
			c_argv[0] = string;
			}
		if (chdir(c_argv[0])) sysemsg(c_argv[0]);
	} else if (!strncmp("write", cmd, len) && cmdlng(5, 1)) {
		if (c_argc > 1) { emsg(ambigous); return; }
		if (name == NULL && c_argc != 0) name = strsave(c_argv[0]);
		if (force) {
			if (c_argc == 0) {
				if (P(P_RO)) {
					sprintf(string, "\"%s\" File is read only", name);
					emsg(string);
					return;
				} else save(name);
			} else {
				if (!stat(c_argv[0], &buf)) {
					sprintf(string, "File exists@- use \"w! %s\" to overwrite",
						c_argv[0]);
					emsg(string);
					return; }
				save(c_argv[0]);
			}
		} else {
			if (c_argc == 0)	save(name);
			else				save(c_argv[0]);
		}
	} else if (!strncmp("wq", cmd, len) && cmdlng(2, 2)) {
		if (c_argc > 1) { emsg(ambigous); return; }
        if (name == NULL && c_argc != 0) name = strsave(c_argv[0]);
		if (force) {
			if (c_argc == 0)	ok = save(name);
			else				ok = save(c_argv[0]);
		} else {
			if (c_argc == 0) {
				if (P(P_RO)) {
					sprintf(string, "\"%s\" File is read only", name);
					emsg(string);
					return;
				} else ok = save(name);
			} else {
				if (!stat(c_argv[0], &buf)) {
					sprintf(string, "File exists@- use \"wq! %s\" to overwrite",
						c_argv[0]);
					emsg(string);
					return; }
				ok = save(c_argv[0]);
			}
		}
		if (ok) quit();
	} else if (!strncmp("edit", cmd, len) && cmdlng(4, 1)) {
		if (c_argc > 1) { emsg(ambigous); return; }
		doecmd(c_argv[0], force);
	/*
	 * The command ":e#" gets expanded to something like ":efile", so
	 * detect that case here.
	 */
	} else if (*cmd == 'e' && c_argc == 0) {
		if (cmd[1] == '!')
			(void) doecmd(&cmd[2], TRUE);
		else
			(void) doecmd(&cmd[1], FALSE);
	} else if (!strncmp("file", cmd, len) && cmdlng(4, 1)) {
		if (c_argc != 0) name = c_argv[0];
		fileinfo(name);
	} else if (!strncmp("read", cmd, len) && cmdlng(4, 1)) {
		/* read reads a file at EOF */
		if (c_argc > 1) { emsg(ambigous); return; }
		if (c_argc == 0) c_argv[0] = name;
		if (!addfile(c_argv[0])) {
			edits = TRUNC;
			undosize = filesize;
			repaint();
		}
	} else if (!strncmp("xit", cmd, len) && cmdlng(3, 1)) {
		if (c_argc != 0) { emsg(extra); return; }
		do_exit();
	} else if (!strncmp("next", cmd, len) && cmdlng(4, 1)) {
		if (c_argc != 0) { emsg(extra); return; }
		if (!force) {
			if (edits) {
				if (P(P_AW)) {
					save(name);
					edits = 0;
				} else {
					sprintf(string, nowrtmsg, "next");
					emsg(string);
					return;
				}
			}
		}
		if ((curfile + 1) < numfiles) {
			if (force)  stuffin(":e! ");
			else		stuffin(":e ");
			stuffin(files[++curfile]);
			stuffin("\n");
		} else
			emsg("No more files@to edit!");
	} else if (!strncmp("rewind", cmd, len) && cmdlng(6, 3)) {
		if (c_argc != 0) { emsg(extra); return; }
		if (numfiles <= 1)		/* nothing to rewind */
			return;
		if (!force) {
			if (edits) {
				if (P(P_AW)) {
					save(name);
					edits = 0;
				} else {
					sprintf(string, nowrtmsg, "rewind");
					emsg(string);
					return;
				}
			}
		}
		curfile = 0;
		if (force)	stuffin(":e! ");
		else		stuffin(":e ");
		stuffin(files[0]);
		stuffin("\n");
	} else if (!strncmp("version", cmd, len) && cmdlng(7, 2)) {
		extern	char	*version;
        if (c_argc != 0) { emsg(extra); return; }
		msg(version);
	} else if (!strncmp("shell", cmd, len) && cmdlng(5, 2)) {
        if (c_argc != 0) { emsg(extra); return; }
		do_shell();
		clear();
		repaint();
		clearstr();
	} else if (!strncmp("quit", cmd, len) && cmdlng(4, 1)) {
		if (c_argc != 0) { emsg(extra); return; }
		if (!force) {
			if (edits) {
				sprintf(string, nowrtmsg, "quit");
				emsg(string);
			} else {
				if ((curfile + 1) < numfiles) {
					sprintf(string, "%d %s", numfiles - curfile - 1, morefiles);
					emsg(string);
				} else
					quit();
			}
		} else quit();
	} else {
		sprintf(string, "What?|%s: Not an editor command", cmd);
		emsg(string);
	}
}


int
addfile(fname)
	char	*fname;
{
	int		fd;
	long	oldsize;

	if (stat(fname, &buf)) {
		sysemsg(fname);
		return 1;	}
	if ((fd = open(fname, O_RDONLY)) == -1) {
		sysemsg(fname);
		return 1;	}
	oldsize = filesize;
	if (enlarge(buf.st_size)) return 1;
	if (read(fd, mem + filesize, buf.st_size) == -1) {
		sysemsg(fname);
		return 1;	}
	filesize += buf.st_size;
	maxpos = mem + filesize;
	close(fd);
	setpage(mem + oldsize);

	return 0;
}


void
do_exit()
{
	if (edits) {
			if (!save(name)) return;
		}
	if ((curfile + 1) < numfiles) {
		sprintf(string, "%d %s", numfiles - curfile - 1, morefiles);
		emsg(string);
	} else
		quit();
}


int
doecmd(arg, force)
	char	*arg;
	int		force;
{
	if (!force && edits) {
		sprintf(string, nowrtmsg, "edit");
		emsg(string);
		if (altfile)
			free(altfile);
		altfile = strsave(arg);
		return FALSE;
	}
	if (arg != NULL) {
		if (name != NULL && strcmp(arg, name) == 0) {
			if (!edits || (edits && !force))
				return TRUE;
		}
		if (altfile) {
			free(altfile);
		}
		altfile = name;
		name = strsave(arg);
	}
	if (name == NULL) {
		emsg("No file|No current filename");
		return FALSE;
	}

	edits = 0;
	filesize = load(name);
	return TRUE;
}


int
getcmdstr(p, x)
	char    *p;
	int		x;
{
	int		c;
	char	*buff, *q;

	buff = p;
	move(maxy, x);
    do {
        switch (c = vgetc()) {
		case CTRL('H'):
		case KEY_BACKSPACE:
		case KEY_LEFT:
            if (p > buff) {
                p--;
				move(maxy, x);
                for (q = buff; q < p ;q++)
                    addch(*q);
            } else {
				*buff = '\0';
                msg("");
                return 1;
            }
            break;
        case ESC:           /* abandon command */
			*buff = '\0';
            msg("");
            return  1;
        case NL:            /* done reading the line */
        case CR:
            break;
        default:        /* a normal character */
            addch(c);
            *p++ = c;
            break;
        }
		refresh();
    } while (c != NL && c != CR);

    *p = '\0';
	return 0;
}


void
clearstr()
{
	int n;

	move(maxy, 0);
	for (n = 0; n < maxx; n++) addch(' ');
	move(maxy, 0);
}


void
emsg(s)		/**** displays an error message *****/
	char	*s;
{
	int	cnt;

	if (P(P_EB)) beep();
	clearstr();
	attrset(A_REVERSE);
	cnt = outmsg(s);
	attrset(A_NORMAL);
	if (cnt >= (maxx - 25)) {		/* 25 = status */
		addch('\n');
		wait_return(); }
}


void
sysemsg(s)		/*** System error message *****/
	char	*s;
{
	char	string[256];

	sprintf(string, "%s: %s", s, sys_errlist[errno]);
	emsg(string);
}


void
smsg(s)		/*** displays mode if showmode set *****/
	char	*s;
{
	if (P(P_MO)) {
		msg(s);
		setcur();
		}
}


void
msg(s)	/************* displays s on status line *****************/
char	*s;
{
	clearstr();
	if (outmsg(s) >= (maxx - 25)) {		/* 25 = status */
		addch('\n');
		wait_return(); }
}


outmsg(s)
	char	*s;
{
	char	*poi;
	int		cnt = 0;

	move(maxy, 0);
	poi = strchr(s, '|');

	if (P(P_TE)) {
		poi = s;
		while (*poi != '\0' && *poi != '@' && *poi != '|') {
			addch(*poi++);
			cnt++; }
	} else {
    	if (poi) poi++;
			else poi = s;
		while (*poi) {
			if (*poi == '@') addch(' ');
				else addch(*poi);
			poi++;
			cnt++;
		}
	}
	return cnt;
}



void
wait_return()
{
	register char	c;

	clearstr();
	attrset(A_REVERSE);
	mvaddstr(maxy, 0, "[Hit return to continue]");
	attrset(A_NORMAL);
	refresh();
	do {
		c = getch();
	} while (c != CR && c != NL && c != ' ' && c != ':');
	clear();
	repaint();
	clearstr();
}
