/* DOSIO.C - I/O for MSDOS - BVI
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * Copyright 1996 by Gerhard Buergmann (G.Buergmann@nl.cis.philips.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */

#include <stdio.h>
#include <conio.h>
#include <sys/stat.h>
#include <errno.h>
#include <alloc.h>
#include <fcntl.h>
#include <bios.h>
#include <io.h>
#include <string.h>
#include <stdlib.h>
#include "doscur.h"
#include "bvi.h"
#include "set.h"

char	*terminal = "ansi";

extern maxy;
struct WINDOW scr;

int	stdscr = 0;

int	COLS = 80;
int	LINES =	25;
int	ECHO = TRUE;
int	NODEL = FALSE;

static  struct  stat    buf;

int	inch(void);

int
attrset(int attr)
{
	switch(attr) {
	case A_NORMAL:  textcolor(P(P_CO) & 0x07);
					textbackground((P(P_CO) & 0xf0) >> 4);
					break;
	case A_BOLD:	textcolor((P(P_CO) & 0x07) | 0x08);
					textbackground((P(P_CO) & 0xf0) >> 4);
					break;
	case A_REVERSE: textcolor((P(P_CO) & 0xf0) >> 4);
					textbackground(P(P_CO) & 0x0f);
					break;
	case A_BLINK:	textcolor((P(P_CO) & 0xf0) >> 4);
					textbackground(P(P_CO) & 0x0f);
					break;
	}
}


int
getch()
{
	int	x;

	if (NODEL) if (!kbhit()) return(ERR);
	x = bioskey(0);
	if ((x & 0xff) == 0) return(x);
		else {
			x &= 0xff;
			if (ECHO) putch(x);
			return(x); }
}


void
delch()
{
	int x, y;

	x = wherex(); y = wherey();
	movetext(x + 1, y, COLS - x, y, x, y);
	gotoxy(COLS, y);
	putch(' ');
	gotoxy(x, y);
}


void
insch(int c)
{
	char line[80];
	int x, y;

	x = wherex(); y = wherey();
	gettext(x, y, COLS - x, y, line);
	puttext(x + 1, y, COLS - x + 1, y, line);
	putch(c);
}


int
inch()
{
	return peek(0xb800, (wherex() - 1) + (wherey() - 1) * COLS);
}

int
mvinch(y, x)
{
	gotoxy(x + 1, y + 1);
	return inch();
}

int
getstr(p)
	char	*p;
{
	int	c;
	char	*buff, *q;

	buff = p;
	do {
		switch (c = vgetc()) {

		default:		/* a normal character */
			addch(c);
			*p++ = c;
			break;

		case BS:
			if (p > buff) {
				p--;
				move(maxy, 1);
				for (q = buff; q < p ;q++)
					addch(*q);
			} else {
				msg("");
				return NULL;		/* back to cmd mode */
			}
			break;

		case '@':			/* line kill */
			p = buff;
			break;

		case ESC:			/* abandon command */
			msg("");
			return  NULL;

		case NL:			/* done reading the line */
		case CR:
			break;
		}
	} while (c != NL && c != CR);

	*p = '\0';
	return(c);
}


int
save(fname)   /*********** Speichern des gepatchten Files ********************/
	char	*fname;
{
	int		fd;
	char	string[255];
	char	*newstr;
	long	written;
	unsigned n, to_write;

	if (!fname) {
		emsg("No file|No current filename");
		return 0;
		}
	if (stat(fname, &buf) == -1) newstr = "[New file] ";
		else newstr = "";

	if ((fd = open(fname, O_WRONLY|O_CREAT|O_BINARY, S_IREAD|S_IWRITE)) < 0) {
		sysemsg(fname);
		return 0;
		}
	written = 0L;
	do {
		to_write = (filesize - written) > 0xfffe ? 0xfffe : (filesize - written);
		if ((n = write(fd, mem + written, to_write)) == 0xffff) {
			sysemsg(fname);
			close(fd);
			return(0L);
			}
		written += (unsigned long)n;
	} while (written < filesize);
	close(fd);
	edits = 0;
	sprintf(string, "\"%s\" %s%ld@bytes", fname, newstr, filesize);
	msg(string);
	return 1;
}


long
load(fname)
	char	*fname;
{
	int		fd;
	char	string[MAXCMD];
	unsigned chunk, n;

	if ((fd = open(fname, O_RDONLY|O_BINARY)) < 0) {
		new = 1;
		buf.st_size = 0L;
	} else {
		if (stat(fname, &buf) == -1) {
			sysemsg(fname);
			close(fd);
			return(0L);
			}
		new = 0;
		if (access(fname, 2)) {
			P(P_RO) = TRUE;
			params[P_RO].flags |= P_CHANGED;
			}
	}
	filesize = 0L; /** buf.st_size; **/
	memsize = buf.st_size + 100L;

	if (mem != NULL) farfree(mem);
	clear_marks();
	if ((mem = (char *)farmalloc(memsize)) == NULL) {
		emsg("Out of memory");
		close(fd);
		return(0L);
		}
	if (!new) {
		chunk = buf.st_size > 0xfffe ? 0xfffe : buf.st_size;
		do {
			if ((n = read(fd, mem + filesize, chunk)) == 0xffff) {
				sysemsg(fname);
				close(fd);
				return(0L);
				}
			filesize += (unsigned long)n;
		} while (filesize < buf.st_size);
		close(fd);
	}
	pagepos = mem;
	maxpos = (PTR)(mem + buf.st_size);
	if (new)
		sprintf(string, "\"%s\" [New File]", name);
	else
		sprintf(string, "\"%s\", %ld bytes", fname, filesize);
	if (name) msg(string);
	loc = HEX;
	x = 8; y = 0;
	repaint();
	return(buf.st_size);
}


void
bvi_init(dir)
	char	*dir;
{
	char	*poi;
	char	*initstr;
	char	rcpath[255];

	shell = getenv("COMSPEC");
	if (shell == NULL || *shell == '\0')
		shell = "COMMAND.COM";

	strcpy(rcpath, dir);
	poi = strrchr(rcpath, '\\');
	*poi = '\0';
	strcat(rcpath, "\\BVI.RC");


	if ((initstr = getenv("BVIINIT")) != NULL) {
			docmdline(initstr);
	}
	read_rc("BVI.RC");
	read_rc(rcpath);
}


int
enlarge(add)
	long	add;
{
	PTR		newmem;
	long	savecur, savepag, savemax;

	savecur = curpos - mem;
	savepag = pagepos - mem;
	savemax = maxpos - mem;

	newmem = (PTR)farrealloc(mem, memsize + add);
	if (newmem == NULL) {
		emsg("Out of memory");
		return 1; }

	mem = newmem;
	memsize += add;
	curpos = mem + savecur;
	pagepos = mem + savepag;
	maxpos = mem + savemax;
	current = curpos + 1;

	return 0;
}


void
do_shell()
{
	system("");
}
