/* IO.C - file in/out for BVI
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * Copyright 1996 by Gerhard Buergmann (G.Buergmann@nl.cis.philips.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <curses.h>
#include "bvi.h"
#include "set.h"

static	struct	stat	buf;
char	*terminal;

int
save(fname)   /*********** Speichern des gepatchten Files ********************/
	char	*fname;
{
	int		fd;
	char	string[255];
	char	*newstr;

	if (!fname) {
		emsg("No current filename");
		return 0;
		}
	if (stat(fname, &buf) == -1) newstr = "[New file] ";
		else newstr = "";

	if ((fd = open(fname, O_WRONLY|O_CREAT, S_IREAD|S_IWRITE)) < 0) {
        sysemsg(fname);
		return 0;
		}
	if (write(fd, mem, filesize) != filesize) {
		sysemsg(fname);
		close(fd);
		return 0;
		}
	close(fd);
	edits = 0;
	sprintf(string, "\"%s\" %s%d bytes", fname, newstr, filesize);
	msg(string);
	return 1;
}


long
load(fname)
	char	*fname;
{
	int		fd;
	char	string[MAXCMD];


	new = 1;
	buf.st_size = 0L;
	if (fname != NULL) {
		if ((fd = open (fname, O_RDONLY)) > 0) {
			if (stat(fname, &buf) == -1) {
				sysemsg(fname);
				close(fd);
				return(0L);
				}
			new = 0;
			if (access(fname, W_OK)) {
				P(P_RO) = TRUE;
				params[P_RO].flags |= P_CHANGED;
				}
		}
	}
	memsize = buf.st_size + 1000L;
	if (mem != NULL) free(mem);
	if ((mem = (char *)malloc(memsize)) == NULL) {
		emsg("Out of memory");
		close(fd);
		return(0L);
		}
	clear_marks();
	if (!new) {
		if (read(fd, mem, buf.st_size) != buf.st_size) {
            sysemsg(fname);
			close(fd);
			return(0L);
		}
		close(fd);
	}
	pagepos = mem;
	maxpos = mem + buf.st_size;
	filesize = buf.st_size;
	if (fname != NULL) {
		if (new)
			sprintf(string, "\"%s\" [New File]", fname);
		else
			sprintf(string, "\"%s\" %s%d bytes", fname,
				P(P_RO) ? "[Read only] " : "",
				filesize);
		msg(string);
	}
	loc = HEX;
	x = 8; y = 0;
	repaint();
	return(buf.st_size);
}


void
bvi_init()
{
	char    *initstr;
	char    rcpath[255];

	terminal = getenv("TERM");
	shell = getenv("SHELL");
	if (shell == NULL || *shell == '\0')
		shell = "/bin/sh";

	if ((initstr = getenv("BVIINIT")) != NULL) {
			docmdline(initstr);
			return; }

	strcpy(rcpath, getenv("HOME"));
	strcat(rcpath, "/.bvirc");
	if (stat(rcpath, &buf) == 0) {
		if (buf.st_uid == getuid()) read_rc(rcpath);
		}

	strcpy(rcpath, ".bvirc");
	if (stat(rcpath, &buf) == 0) {
		if (buf.st_uid == getuid())	read_rc(rcpath);
		}
}


int
enlarge(add)
	long	add;
{
	char	*newmem;
	long	savecur, savepag, savemax;

	savecur = curpos - mem;
	savepag = pagepos - mem;
	savemax = maxpos - mem;

	newmem = realloc(mem, memsize + add);
	if (newmem == NULL) {
		emsg("Out of memory");
		return 1; }

	mem = newmem;
	memsize += add;
	curpos = mem + savecur;
	pagepos = mem + savepag;
	maxpos = mem + savemax;
	current = curpos + 1;

	return 0;
}


void
do_shell()
{
	addch('\n');
	savetty();
	system(shell);
	resetty();
}


char *
strdup(s)
	char	*s;
{
        char    *p;
        unsigned n;

        n = strlen(s) + 1;
        if ((p = (char *)malloc(n)) != NULL)
                memcpy(p, s, n);
        return (p);
}

