/* re - regular expression search for bvi
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * 1996-01-06 created;
 *
 * Copyright 1996 by Gerhard Buergmann (G.Buergmann@nl.cis.philips.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */

/* You cannot use a common regexp subroutine, because \0 is a regular
 * character in a string !
 */
 
#include	<string.h>
#include	<ctype.h>

#ifdef __MSDOS__
#	include "doscur.h"
#else
#	include <curses.h>
#endif

#include	"bvi.h"
#include	"set.h"

#define END		0
#define ONE		1
#define STAR	2

static	char	smem[BUFFER], *poi;
static	int		smode;

static	int sbracket();

char	act_pat[MAXCMD];  /* found pattern */
char	pattern[MAXCMD + 1];
char	*notfound = "Fail|Pattern not found";
char	*emptyclass = "Bad character class|Empty byte class '[]' or '[^]' cannot match";


/*
 *  Compiling an ASCII sequence to a regex string
 */
int
ascii_comp(pattern)
	char	*pattern;
{
	char	*end;
	char	*comp;
	char	cc, cc1;
	char	*counter;
	int		count;
	int		bracket, dot;

	comp = smem;
	poi = pattern;
	while (*poi != END) {
		bracket = FALSE;
		if (P(P_MA)) {
			if (*poi == '[') bracket = TRUE;
		} else {
			if (*poi == '\\' && *(poi + 1) == '[') {
				bracket = TRUE;
				*poi++; }
		}
		if (bracket) {
			if (!(end = strchr(poi, ']'))) {
				emsg("Missing ]");
				return 1; }
			poi++;
			if (*poi == ']' || (*poi == '^' && *(poi + 1) == ']')) {
				emsg(emptyclass);
				return 1; }
			if (P(P_MA)) {
				if (*(end + 1) == '*') *comp++ = STAR;
					else	*comp++ = ONE;
			} else {
				if (*(end + 1) == '\\' && *(end + 2) == '*') *comp++ = STAR;
					else	*comp++ = ONE;
			}
			count = 0;
			counter = comp;
			*comp++;
			if (*poi != '^') {
				*comp++ = '\0';
				count++; }

			while (end > poi) {
				if (*poi == '-') {
					if (P(P_IC)) {
						cc = toupper(*(poi - 1));
						cc1 = toupper(*(poi + 1));
					} else {
						cc = *(poi - 1);
						cc1 = *(poi + 1);
					}
					while (cc <= cc1) {
						*comp++ = cc++;
						count++;
						}
					*poi++; *poi++;
				} else {
					count++;
					if (P(P_IC))
						*comp++ = toupper(*poi++);
					else
						*comp++ = *poi++;
				}
			}
			*poi++;
			*counter = count;
		} else {
			dot = FALSE;
			if (P(P_MA)) {
				if (*poi == '.') dot = TRUE;
			} else {
				if (*poi == '\\' && *(poi + 1) == '.') {
					dot = TRUE; *poi++; }
			}

			if (*poi == '\\') {
				switch (*(poi + 1)) {
				case 'n': *++poi = '\n'; break;
				case 'r': *++poi = '\r'; break;
				case 't': *++poi = '\t'; break;
				case '0': *++poi = '\0'; break;
				default : *++poi;
				}
			}
			if (P(P_MA))
				if (*(poi + 1) == '*') *comp++ = STAR;
					else	*comp++ = ONE;
			else
				if (*(poi + 1) == '\\' && *(poi + 2) == '*') *comp++ = STAR;
					else    *comp++ = ONE;

			if (dot) {
				*comp++ = 0;
				*poi++;
			} else {
				*comp++ = 1;
				if (P(P_IC))
					*comp++ = toupper(*poi++);
				else
					*comp++ = *poi++;
			}
		}
		if (P(P_MA)) {
			if (*poi == '*') poi++;
		} else {
			if (*poi == '\\' && *(poi + 1) == '*') { poi++; poi++; }
		}
	}
	*comp = END;
	smode = ASCII;
	return 0;
}


/*
 * Compiling a hex expression to a regex string
 */
int
hex_comp(pattern)
	char	*pattern;
{
	char	*end;
	char	*comp;
	int		cc, ccm;
	char	*counter;
	int		count, nr;

	comp = smem;
	poi = pattern;
	while (*poi != END) {
		while (*poi == ' ' || *poi == '\t') poi++;
		if (*poi == '[') {
			if (!(end = strchr(poi, ']'))) {
				emsg("Missing ]");
				return 1; }
			poi++;
			while (*poi == ' ' || *poi == '\t') poi++;

			if (*poi == ']' || (*poi == '^' && *(poi + 1) == ']')) {
				emsg(emptyclass);
				return 1; }
			if (*(end + 1) == '*') *comp++ = STAR;
				else	*comp++ = ONE;
			count = 1;
			counter = comp;
			*comp++;
			if (*poi == '^')
				*comp++ = *poi++;
			else
				*comp++ = '\0';
			while (end > poi) {
				if (*poi == ' ' || *poi == '\t') poi++;
				else if (*poi == '-') {
					cc = *(comp - 1);
					*poi++;
					if ((ccm = hexchar()) < 0) return 1;
					while (cc <= ccm) {
						*comp++ = cc++;
						count++;
						}
					*poi++; *poi++;
				} else {
					if ((nr = hexchar()) < 0) return 1;
					count++;
					*comp++ = nr;
				}
			}
			*poi++;
			*counter = count;
			if (*poi == '*') *poi++;
		} else {
			if (*poi == '.') {
				if (*(poi + 1) == '*') { *comp++ = STAR; *poi++; }
					else	*comp++ = ONE;
				*comp++ = 0;
				*poi++;
			} else {
				if ((nr = hexchar()) < 0) return 1;
				if (*poi == '*') { *poi++; *comp++ = STAR; }
					else    *comp++ = ONE;
				*comp++ = 1;
				*comp++ = nr;
			}
		}
	}
	*comp = END;
	smode = HEX;
	return 0;
}


int
hexchar()
{
	int		nr;
	char	tmpbuf[3];
	char	string[55];

	if (isxdigit(*poi)) {
		tmpbuf[0] = *poi++;
		tmpbuf[1] = '\0';
		if (isxdigit(*poi)) tmpbuf[1] = *poi++;
		tmpbuf[2] = '\0';
		sscanf(tmpbuf, "%2x", &nr);
        while (*poi == ' ' || *poi == '\t') poi++;
		return nr;
	} else {
		emsg("Bad hex character@in expression");
		return -1;
	}
}


int
regexec(start, scan)
	PTR		start;
	char	*scan;
{
	char	*act;
	int		count, test;

	/*
	scan = smem;
	*/
	act = act_pat;
		while (*scan != 0) {
		switch (*scan++) {
		case ONE:	/* exactly one character */
					count = *scan++;
					if (P(P_IC) && smode == ASCII)	test = toupper(*start);
						else 						test = *start;
					if (count == 1) {
						if (test != *scan) return 0;
						scan++;
					} else if (count > 1) {
						if (sbracket(test, scan, count)) return 0;
						scan += count;
					}
					*act++ = *start++;
					break;
		case STAR:  /* zero or more characters */
					count = *scan++;
					if (P(P_IC) && smode == ASCII)	test = toupper(*start);
						else 						test = *start;
					if (count == 1) {
						while(test == *scan) {
							*act++ = *start++;
							if (P(P_IC) && smode == ASCII)	test = toupper(*start);
								else 						test = *start;
							}
						scan++;
					} else if (count > 1) {
						if (*scan == '^') {
							while (start < maxpos) {
								if (sbracket(test, scan, count)) return 0;
								if (regexec(start, scan)) { *act = '\0'; return 1; }
								start++;
							}
						} else {
							while(!sbracket(test, scan, count)) {
								*act++ = *start++;
								if (P(P_IC) && smode == ASCII)	test = toupper(*start);
									else 						test = *start;
								}
							scan += count;
						}
					} else {
						while (start < maxpos) {
							if (regexec(start, scan)) { *act = '\0'; return 1; }
							start++;
							}
					}
		}
	}
	*act = '\0';
	return 1;	/* found */
}


static int
sbracket(start, scan, count)
	int		start;
	char	*scan;
	int		count;
{
	if (*scan++ == '^') {
		if (!memchr(scan, start, --count)) return 0;
	} else {
		if (memchr(scan, start, --count)) return 0;
	}
	return 1;
}


PTR
end_word(start)
	PTR	start;
{
	PTR	pos;

	pos = start;
	if (!isprint(*pos & 0xff)) return start;
	while (isprint(*pos & 0xff)) if (pos++ > maxpos) return start;
	return --pos;
}


/* wordsearch serves the 'W' and 'w' - command
 */
PTR
wordsearch(start, mode)
	PTR		start;
	char	mode;
{
	PTR		found;
	PTR		pos;
	int		ccount;

	pos = start + 1;
	do {
		while (isprint(*pos & 0xff)) if (pos++ > maxpos) return start;
		while (!isprint(*pos & 0xff)) if (pos++ > maxpos) return start;
		found = pos;
		ccount = 0;
		while (isprint(*pos & 0xff)) {
			if (pos++ > maxpos) return start;
			ccount++; }
		if (ccount < P(P_WL)) continue;

		if (mode == 'W') {
			if (*pos == '\0' || *pos == '\n') return found;
		} else {
			return found;
		}
	} while (pos < maxpos);
	return	start;
}


/* backsearch serves the 'b' and 'B' command
 */
PTR
backsearch(start, mode)
	PTR		start;
	char	mode;
{
	PTR		pos;
	int		ccount;

	pos = start - 1;
	do {
		if (mode == 'B') {
			while (*pos != '\0' && *pos != '\n') if (pos-- < mem) return start;
		} else {
			while (!isprint(*pos & 0xff))		if (pos-- < mem) return start;
		}
		pos--;
		ccount = 0;
		while (isprint(*pos & 0xff)) {
			if (pos-- < mem) return start;
			ccount++; }
		if (ccount >= P(P_WL)) 	return (pos + 1);

	} while (pos > mem);
	return	start;
}


do_sub(str)
	char	*str;
{

}


/*  The FIND and REPLACE routine
 *	flag = SEARCH(0) or REPLACE(1)
 */
int
searching(ch, line, flag)
	int		ch;
	char	*line;
	int		flag;
{
	PTR		actpos;
	PTR		startpos;
	PTR		endpos;
	int		ch1, n;
	char	*found;
	char	*cmd;
	int		sdir;
	static	int	direct;
	static	int	again = 0;

	if ((strchr("\\%", ch) && loc == ASCII)
		|| (strchr("/?", ch) && loc == HEX))
				toggle();

	if (!strchr("Nn", ch)) {
		switch (ch) {
		case 'F':
		case 'f':
			direct = (ch == 'f' ? FORWARD : BACKWARD);
			pattern[0] = vgetc();
			pattern[1] = '\0';
			if (ascii_comp(pattern)) return;
			break;
		case '/':
		case '?':
			direct = (ch == '/' ? FORWARD : BACKWARD);
			cmd = patcpy(pattern, line, ch);
			if (pattern[0] == '\0') break;
			if (ascii_comp(pattern)) return;
			again = 1;
			break;
		case '%':
		case '\\':
			direct = (ch == '\\' ? FORWARD : BACKWARD);
			cmd = patcpy(pattern, line, ch);
			if (hex_comp(pattern)) break;
			again = 1;
			break;
		}
		if (!again) return 1;
	} else {
		cmd = "";
		if (again) {
			msg("/");
		} else {
			emsg("No previous expression");
			return(0); }
	}
	move(maxy, 0);
	refresh();
	sdir = (ch == 'N') ? !direct : direct;

#ifdef DEBUG
	sprintf(string, "%02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x",
		smem[0], smem[1], smem[2], smem[3], smem[4], smem[5], smem[6],
		smem[7], smem[8], smem[9], smem[10], smem[11]);
	msg(string);
	getch();
#endif

	startpos = pagepos + y * Anzahl + xpos();
	endpos = maxpos - 1;
	if (sdir == FORWARD) {
		found = fsearch(startpos + 1, endpos);
		if (!found && flag == SEARCH)	 /** REPLACE only to EOF **/
			if (P(P_WS)) {
				msg("Search wrapped BOTTOM|Search wrapped around BOTTOM of buffer");
				found = fsearch(mem, startpos);
				}
	} else {
		found = rsearch(startpos - 1, mem);
		if (!found && flag == SEARCH)
			if (P(P_WS)) {
				msg("Search wrapped TOP|Search wrapped around TOP of buffer");
				found = rsearch(endpos, startpos);
				}
	}
	if (!found) {
		if (!flag) {
			if (P(P_WS)) {
				emsg(notfound);
			} else {
				if (P(P_TE))
				  sprintf(string, "No match to %s", sdir == FORWARD ? "BOTTOM" : "TOP");
				else
				  sprintf(string, "Address search hit %s without matching pattern",
						sdir == FORWARD ? "BOTTOM" : "TOP");
				emsg(string);
			}
		}
	} else {
		setpage(found);
		if (flag == REPLACE) {
			cmd = patcpy(pattern, cmd, ch);
			if (!strncmp(cmd, "gc", 2)) {	/** conditional **/
				repaint();
				msg("Replace?");
				move(y, x);
				if (vgetc() != 'y') goto SKIP;
			}
			repl_count++;
			current_start = pagepos + y * Anzahl + xpos();
			if (loc == ASCII) {
				undo_count = strlen(pattern);
				for (n = 0; n < undo_count; n++) {
					undo_buf[n] = current_start[n];
					current_start[n] = pattern[n];
					}
			} else {
				poi = pattern;
				undo_count = 0;
				while (*poi) {
					if (*poi == ' ' || *poi == '\t') poi++;
					else {
						if ((n = hexchar()) < 0) {
							emsg("Badly formed replacement pattern");
							return 1; }
						undo_buf[undo_count] = current_start[undo_count];
						current_start[undo_count] = n;
						undo_count++;
					}
				}
			}
			undo_start = current_start;
			edits = EDIT;
SKIP:
			if (*cmd == 'g') searching(ch, line, REPLACE);
		} else {
			if (cmd) {
				switch(*cmd) {
				case 'z':	do_z(*++cmd);
							break;
				case 's':	do_sub(*++cmd);
							break;
				case ';':	searching(ch, cmd + 2, SEARCH);
				case '\0':	break;
				default:  	beep();
				}
			}
		}
	}
	return(0);
}


char *
patcpy(s1, s2, delim)
	register	char	*s1, *s2;
	register	char	delim;
{
	while (*s2 != '\0' && *s2 != delim) {
		if (*s2 == '\\' && *(s2 + 1) == delim) *s2++;
		*s1++ = *s2++;
		}
	*s1 = '\0';
	if (*s2 != '\0') return ++s2;
	return '\0';
}


PTR
fsearch(start, end)
	PTR start;
	PTR end;
{
	PTR	spos;

	for (spos = start; spos <= end; spos++) {
		if (regexec(spos, smem)) return(spos);
		}
	return(NULL);
}


PTR
rsearch(start, end)
	PTR start;
	PTR end;
{
	PTR	spos;

	for (spos = start; spos >= end; spos--) {
		if (regexec(spos, smem)) return(spos);
		}
	return(NULL);
}
