/* SET.C - performing :set - command
 *
 * NOTE: Edit this file with tabstop=4 !
 *
 * 1996-02-29 created;
 *
 * Copyright 1996 by Gerhard Buergmann (G.Buergmann@nl.cis.philips.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * See file COPYING for information on distribution conditions.
 */

#include <stdio.h>
#include <string.h>
#include "set.h"
#include "bvi.h"

struct	param	params[] = {
	{ "scroll",		"scroll",	12,		P_NUM },
	{ "window",		"window",	25,		P_NUM },
	{ "offset",		"of",		0,		P_NUM },
	{ "wrapscan",	"ws",		TRUE,	P_BOOL },
	{ "errorbells",	"eb",		FALSE,	P_BOOL },
	{ "showmode",	"mo",		TRUE,	P_BOOL },
	{ "ignorecase",	"ic",		FALSE,	P_BOOL },
	{ "magic",		"ma",		TRUE,	P_BOOL },
	{ "wordlength",	"wl",		4,		P_NUM },
	{ "autowrite",	"aw",		FALSE,	P_BOOL },
	{ "readonly",	"ro",		FALSE,	P_BOOL },
	{ "ustyle",		"us",		FALSE,	P_BOOL },
	{ "terse",		"terse",	FALSE,	P_BOOL },
#ifdef __MSDOS__
	{ "color",		"co",		7,		P_NUM  },
#endif
	{ "",			"",			0,		0, }		/* end marker */

};

extern	char	*terminal;
static	void	showparms();

int
doset(arg)
	char	*arg;		/* parameter string */
{
	register int	i;
	register char	*s;
	int	did_window = FALSE;
	int	state = TRUE;		/* new state of boolean parms. */
	char string[80];

	if (arg == NULL) {
		showparms(FALSE);
		return 0;
	}
	if (!strncmp(arg, "all", 3)) {
		showparms(TRUE);
		return 0;	}
	if (!strncmp(arg, "no", 2)) {
		state = FALSE;
		arg += 2; }
	if (!strncmp(arg, "term", 4)) {
		emsg("Can't change type of terminal from within bvi");
		return 1; }

	for (i = 0; params[i].fullname[0] != '\0'; i++) {
		s = params[i].fullname;
		if (strncmp(arg, s, strlen(s)) == 0)	/* matched full name */
			break;
		s = params[i].shortname;
		if (strncmp(arg, s, strlen(s)) == 0)	/* matched short name */
			break;
	}

	if (params[i].fullname[0] != '\0') {	/* found a match */
		if (params[i].flags & P_NUM) {
			did_window = (i == P_LI);
			if (arg[strlen(s)] != '=' || state == FALSE) {
				sprintf(string, "Option %s is not a toogle", params[i].fullname);
				emsg(string);
				return 1;
			} else {
				params[i].value = atoi(arg + strlen(s) + 1);
				params[i].flags |= P_CHANGED;
#ifdef __MSDOS__
				if (i == P_CO) {
					textcolor(P(P_CO) & 0x07);
					textbackground((P(P_CO) & 0xf0) >> 4);
					clrscr();
					repaint();
					}
#endif
			}
		} else /* boolean */ {
			if (arg[strlen(s)] == '=') {
				emsg("Invalid set of boolean parameter");
				return 1;
			} else {
				params[i].value = state;
				params[i].flags |= P_CHANGED;
			}
		}
	} else  {
		emsg("No such option@- `set all' gives all option values");
		return 1;
	}

	if (did_window) {
		maxy = P(P_LI) - 1;
		new_screen();
	}

	return 0;
}


static	void
showparms(all)
	int	all;	/* show ALL parameters */
{
	register struct	param	*p;
	char	buf[64];

	msg("Parameters:\n");
	for (p = &params[0]; p->fullname[0] != '\0' ;p++) {
		if (!all && ((p->flags & P_CHANGED) == 0))
			continue;
		if (p->flags & P_BOOL)
			sprintf(buf, "        %s%s\n",
				(p->value ? "" : "no"), p->fullname);
		else
			sprintf(buf, "        %s=%d\n", p->fullname, p->value);

		msg(buf);
	}
	sprintf(buf, "        term=%s\n", terminal);
	msg(buf);
	wait_return();
}


read_rc(fn)
	char *fn;
{
	FILE	*fp;
	char	buf[256];

	if((fp = fopen(fn, "r")) == NULL) return -1;
	while(fgets(buf, 255, fp) != NULL) {
				strtok(buf, "\n\r");
				docmdline(buf);
				}
	fclose(fp);
	return 0;
}

